"""BayBE objectives."""

from baybe.objectives.desirability import DesirabilityObjective
from baybe.objectives.single import SingleTargetObjective

__all__ = [
    "SingleTargetObjective",
    "DesirabilityObjective",
]
